import random

import cv2
from PIL import Image, ImageFont, ImageDraw
import numpy as np
from jacksung.utils.data_convert import Coordinate
import os
from importlib import resources
from tqdm import tqdm


def get_pixel_by_coord(img, coord, x, y):
    left, top, x_res, y_res = coord.left, coord.top, coord.x_res, coord.y_res
    if x < left or y > top:
        raise Exception(f'x:{x} or y:{y} is lower than border {left},{top}!'
                        f'left:{left}, top:{top},x_res:{x_res}, y_res:{y_res}.')
    s = img.shape
    if x > left + s[-1] * x_res or y < top - s[-2] * y_res:
        raise Exception(f'x:{x} or y:{y} is greater than border {left + s[-1] * x_res},{top - s[-2] * y_res}!'
                        f'left:{left}, top:{top},x_res:{x_res}, y_res:{y_res}.')
    return img[..., int((top - y) // y_res), int((x - left) // x_res)]


def draw_text(img, xy, font=None, font_size=35, text='test text', color=(0, 0, 0)):
    if xy[0] < 0 or xy[1] < 0:
        xy_txt = (0 if xy[0] < 0 else xy[0], 0 if xy[1] < 0 else xy[1])
        h, w, c = img.shape
        white_block = make_block(h, w, color=(255, 255, 255))
        txt_block = _draw_text(white_block, xy_txt, font, font_size, text, color)
        non_white_mask = np.any(txt_block != [255, 255, 255], axis=-1)
        row_indices, col_indices = np.where(non_white_mask)
        # 3. 计算最大/最小索引
        min_row, max_row = row_indices.min(), row_indices.max()
        min_col, max_col = col_indices.min(), col_indices.max()
        txt_length = max_col - min_col
        txt_width = max_row - min_row
        if xy[0] < 0:
            xy = ((w - txt_length) // 2, xy[1])
        if xy[1] < 0:
            xy = (xy[0], (h - txt_width) // 2)
    return _draw_text(img, xy, font, font_size, text, color)


def _draw_text(img, xy, font=None, font_size=35, text='test text', color=(0, 0, 0)):
    if font is None:
        try:
            with resources.path("jacksung.libs", "times.ttf") as font_path:
                font = ImageFont.truetype(str(font_path), size=font_size)  # 指定字体大小
        except:
            print('load times ttf failed, using the default font')
            font = ImageFont.load_default()
    im = Image.fromarray(img.astype(np.uint8))
    draw = ImageDraw.Draw(im)
    draw.text(xy, text, font=font, fill=color)
    image = np.array(im)
    return image


def _check_border(in_n, n):
    if in_n < 0:
        return 0
    if in_n > n:
        return n
    return in_n


def border(img, point1, point2, color=(0, 0, 255), border=5):
    point1_h, point1_w = point1
    point2_h, point2_w = point2
    h, w, _ = img.shape
    img[_check_border(point1_h - border // 2, h):    _check_border(point1_h + border // 2, h),
    _check_border(point1_w - border // 2, w):    _check_border(point2_w + border // 2, w), :] = color
    img[_check_border(point2_h - border // 2, h):    _check_border(point2_h + border // 2, h),
    _check_border(point1_w - border // 2, w):    _check_border(point2_w + border // 2, w), :] = color
    img[_check_border(point1_h - border // 2, h):    _check_border(point2_h + border // 2, h),
    _check_border(point1_w - border // 2, w):    _check_border(point1_w + border // 2, w), :] = color
    img[_check_border(point1_h - border // 2, h):    _check_border(point2_h + border // 2, h),
    _check_border(point2_w - border // 2, w):    _check_border(point2_w + border // 2, w), :] = color

    return img


def make_block(h, w, color=(255, 255, 255), dtype=np.int32):
    return np.array([[color for _ in range(w)] for _ in range(h)], dtype=dtype)


def get_color_position(value, colors):
    colors_min, colors_max = colors[0][0], colors[-1][0]
    colors = [[(color[0] - colors_min) / (colors_max - colors_min), color[1]] for color in colors]
    i = 0
    while i < len(colors) - 1:
        if value <= colors[i + 1][0]:
            break
        i += 1
    color_str0, color_str1 = colors[i][1], colors[i + 1][1]
    r1, g1, b1, r2, g2, b2 = int(color_str0[1:3], 16), int(color_str0[3:5], 16), int(color_str0[5:7], 16), \
        int(color_str1[1:3], 16), int(color_str1[3:5], 16), int(color_str1[5:7], 16)
    r = (value - colors[i][0]) / (colors[i + 1][0] - colors[i][0]) * (r2 - r1) + r1
    g = (value - colors[i][0]) / (colors[i + 1][0] - colors[i][0]) * (g2 - g1) + g1
    b = (value - colors[i][0]) / (colors[i + 1][0] - colors[i][0]) * (b2 - b1) + b1
    return np.array((b, g, r))


def make_color_map(colors, h, w, unit=''):
    colors_map = np.zeros((h, w, 3), dtype=np.uint8) + 255
    l_margin, r_margin = 300, 200
    w = w - l_margin - r_margin
    for i in range(l_margin, w + l_margin):
        i = i - l_margin
        colors_map[:100, i + l_margin] = get_color_position(i / w, colors)
        if i in [0, w // 2, w - 1]:
            text = str(round((i / w) * (colors[-1][0] - colors[0][0]) + colors[0][0]))
            if i == 0:
                text += unit
            try:
                with resources.path("jacksung.libs", "times.ttf") as font_path:
                    font = ImageFont.truetype(str(font_path), size=150)  # 指定字体大小
            except:
                print('load times ttf failed, using the default font')
                font = ImageFont.load_default()
            colors_map = draw_text(colors_map, (i - 100 + l_margin, 100),
                                   font=font, text=text)
    return colors_map


def crop_png(input_path, left=0, top=0, right=None, bottom=None, right_margin=0, bottom_margin=0):
    # 打开 PNG 图像
    image = Image.open(input_path)
    width, height = image.size
    if right is None:
        right = width
    if bottom is None:
        bottom = height
    right -= right_margin
    bottom -= bottom_margin
    # 对图像进行裁剪
    cropped_image = image.crop((left, top, right, bottom))
    # 保存裁剪后的图像
    cropped_image.save(input_path)


def concatenate_images(imgs, direction="h"):
    # 读取所有的图片
    # 获取图片的宽度和高度
    heights = [img.shape[0] for img in imgs]
    widths = [img.shape[1] for img in imgs]

    if direction == "h":
        # 水平拼接，高度不变，宽度为所有图片宽度之和
        max_height = max(heights)
        total_width = sum(widths)
        if imgs[0].ndim == 2:
            new_img = np.zeros((max_height, total_width), dtype=np.uint8)
        else:
            new_img = np.zeros((max_height, total_width, 3), dtype=np.uint8)
        x_offset = 0
        for img in imgs:
            new_img[:, x_offset:x_offset + img.shape[1]] = img
            x_offset += img.shape[1]
    elif direction == "v":
        # 垂直拼接，宽度不变，高度为所有图片高度之和
        max_width = max(widths)
        total_height = sum(heights)
        if imgs[0].ndim == 2:
            new_img = np.zeros((total_height, max_width), dtype=np.uint8)
        else:
            new_img = np.zeros((total_height, max_width, 3), dtype=np.uint8)
        y_offset = 0
        for img in imgs:
            new_img[y_offset:y_offset + img.shape[0], :] = img
            y_offset += img.shape[0]
    else:
        raise ValueError("Invalid direction. Please choose 'h' or 'v'.")
    return new_img


def create_gif(images_in, output_path, duration=500, idx=None):
    images = []
    if type(images_in) == str:
        for file_name in sorted(os.listdir(images_in)):
            # if file_name.endswith('.png'):
            file_path = os.path.join(images_in, file_name)
            images.append(Image.open(file_path))
    else:
        for img in images_in:
            images.append(Image.fromarray(cv2.cvtColor(img.astype(np.uint8), cv2.COLOR_BGR2RGB)))
    if idx is not None:
        images = images[idx[0]:idx[1]]
    images[0].save(output_path, save_all=True, append_images=images[1:], duration=duration, loop=0)


def zoom_image(image, scale_factor=2):
    # 获取图像的尺寸
    height, width = image.shape[:2]
    # 计算缩放后的尺寸
    new_dimensions = (int(width * scale_factor), int(height * scale_factor))
    # 使用cv2.resize进行缩放
    zoomed_image = cv2.resize(image, new_dimensions, interpolation=cv2.INTER_LINEAR)
    return zoomed_image


def zoomAndDock(img, zoom_rectangle, docker, scale_factor=2, border=10):
    corp_png = img[zoom_rectangle[1]:zoom_rectangle[1] + zoom_rectangle[3],
               zoom_rectangle[0]:zoom_rectangle[0] + zoom_rectangle[2], :]
    corp_png = zoom_image(corp_png, scale_factor=scale_factor)
    cv2.rectangle(img, (zoom_rectangle[0], zoom_rectangle[1]),
                  (zoom_rectangle[0] + zoom_rectangle[2], zoom_rectangle[1] + zoom_rectangle[3]), (0, 0, 255), border)
    corp_png[:, 0:border, :] = [0, 0, 0]
    corp_png[:, corp_png.shape[1] - border:corp_png.shape[1], :] = [0, 0, 0]
    corp_png[0:border, :, :] = [0, 0, 0]
    corp_png[corp_png.shape[0] - border:corp_png.shape[0], :, :] = [0, 0, 0]
    img[docker[0]:docker[0] + corp_png.shape[0], docker[1]:docker[1] + corp_png.shape[1], :] = corp_png
    return img


def draw_boundary(img, percent=95):
    h, w = img.shape[:2]
    gradient = np.zeros((h, w))
    for i in range(4, h - 4):
        for j in range(4, w - 4):
            gradient[i, j] += abs(img[i + 1, j] - img[i - 1, j])
            gradient[i, j] += abs(img[i + 2, j] - img[i - 2, j])
            gradient[i, j] += abs(img[i + 3, j] - img[i - 3, j])
            gradient[i, j] += abs(img[i + 4, j] - img[i - 4, j])
            gradient[i, j] += abs(img[i, j + 1] - img[i, j - 1])
            gradient[i, j] += abs(img[i, j + 2] - img[i, j - 2])
            gradient[i, j] += abs(img[i, j + 3] - img[i, j - 3])
            gradient[i, j] += abs(img[i, j + 4] - img[i, j - 4])

    threshold = np.percentile(gradient, percent)
    boundary_mask = gradient > threshold
    return boundary_mask


if __name__ == '__main__':
    pass



import os
import platform
# 从selenium导入webdriver
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.chrome.service import Service
from jacksung.utils.log import oprint as print
from selenium.webdriver.common.by import By


def make_driver(url=None, is_headless=False, tmp_path=None, download_dir=None, options=webdriver.ChromeOptions()):
    options.add_argument("--disable-blink-features=AutomationControlled")
    if tmp_path:
        options.add_argument("crash-dumps-dir=" + tmp_path)
    options.add_argument("--no-sandbox")
    # options.add_argument("--auto-open-devtools-for-tabs")
    options.add_argument('--disable-dev-shm-usage')
    options.add_argument("--disable-web-security")  # 禁用Web安全
    options.add_argument("--allow-running-insecure-content")  # 允许不安全的内容
    options.add_argument('--user-agent=Mozilla/5.0')
    options.add_argument('--ignore-ssl-errors=yes')
    options.add_argument('--allow-insecure-localhost')
    options.add_argument('--ignore-certificate-errors')
    options.add_argument(
        '--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/140.0.0.0 Safari/537.36')  # 设置User-Agent
    options.add_argument("--lang=zh-CN")  # 将语言设置为简体中文，英文为en-US
    options.add_experimental_option("detach", True)
    if download_dir:
        options.add_experimental_option("prefs", {
            "download.default_directory": download_dir,
            "download.prompt_for_download": False,
            "safebrowsing.enabled": True,  # 允许“不安全”文件自动下载
            "safebrowsing.disable_download_protection": True,  # 禁用“可能有害文件”的拦截
            "intl.accept_languages": "zh-CN,zh"
            # "download.directory_upgrade": True,
        })
    options.set_capability('pageLoadStrategy', 'none')
    options.set_capability("unhandledPromptBehavior", "accept")
    if is_headless:
        options.add_argument('--headless')  # 浏览器隐式启动
    # driver_path = os.path.expanduser("~/chrome/chromedriver.exe")
    print('driver is going to init!')
    if platform.system().lower() == 'windows':
        driver_path = None
        # driver_path = os.path.expanduser("~/chrome/chromedriver.exe")
    else:
        driver_path = os.path.expanduser("~/chrome/chromedriver")
    # driver = uc.Chrome(service=Service(driver_path) if driver_path else None, options=options)
    driver = webdriver.Chrome(service=Service(driver_path) if driver_path else None, options=options)
    # driver.maximize_window()
    driver.execute_cdp_cmd("Page.addScriptToEvaluateOnNewDocument", {
        "source": """
            Object.defineProperty(navigator, 'webdriver', {
              get: () => undefined
            })
        """
    })
    driver.implicitly_wait(10)
    driver.set_page_load_timeout(10)
    if url is not None:
        print(f'请求地址：{url}')
        driver.get(url)
    return driver



from jacksung.utils.data_convert import nc2np, np2tif
import numpy as np
import netCDF4 as nc
from einops import rearrange
import os
import shutil
import requests
import time
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
import platform
from jacksung.utils.log import oprint


class Downloader:
    def __init__(self, download_file_path, username, passwd, save_path=None):
        self.download_file_path = download_file_path
        if save_path is not None:
            self.save_path = save_path
        else:
            if platform.system().lower() == 'windows':
                self.save_path = 'D:\\imerg'
            else:
                self.save_path = '/mnt/data1/szj/imerg'
        self.username = username
        self.passwd = passwd

    def make_driver(self, url, is_headless=False, tmp_path=None, download_dir=None):
        options = webdriver.ChromeOptions()
        if tmp_path:
            options.add_argument("crash-dumps-dir=" + tmp_path)
        options.add_argument("--no-sandbox")
        # options.add_argument("--auto-open-devtools-for-tabs")
        options.add_argument('--disable-dev-shm-usage')
        options.add_argument("--disable-web-security")  # 禁用Web安全
        options.add_argument("--allow-running-insecure-content")  # 允许不安全的内容
        options.add_argument('--user-agent=Mozilla/5.0')
        options.add_argument('--ignore-ssl-errors=yes')
        options.add_argument('--allow-insecure-localhost')
        options.add_argument('--ignore-certificate-errors')
        options.add_argument("--lang=zh-CN")  # 将语言设置为简体中文，英文为en-US
        options.add_experimental_option("detach", True)
        if download_dir:
            options.add_experimental_option("prefs", {
                "download.default_directory": download_dir,
                # "download.prompt_for_download": False,
                # "download.directory_upgrade": True,
                # "safebrowsing.enabled": True
            })
        options.set_capability('pageLoadStrategy', 'none')
        options.set_capability("unhandledPromptBehavior", "accept")
        if is_headless:
            options.add_argument('--headless')  # 浏览器隐式启动
        # driver_path = os.path.expanduser("~/chrome/chromedriver.exe")
        print('driver is going to init!')
        if platform.system().lower() == 'windows':
            driver_path = None
            # driver_path = os.path.expanduser("~/chrome/chromedriver.exe")
        else:
            driver_path = os.path.expanduser("~/chrome/chromedriver")
        driver = webdriver.Chrome(service=Service(driver_path) if driver_path else None, options=options)
        # driver.maximize_window()
        driver.implicitly_wait(10)
        driver.set_page_load_timeout(10)
        print('driver is inited!')
        print(f'请求地址：{url}')
        driver.get(url)
        return driver

    # 进度条模块
    def progressbar(self, url, path):
        if not os.path.exists(path):  # 看是否有该文件夹，没有则创建文件夹
            os.mkdir(path)
        start = time.time()  # 下载开始时间
        response = requests.get(url, stream=True)
        # https://data.gesdisc.earthdata.nasa.gov/data/GPM_L3/GPM_3IMERGHH.07/2023/001/3B-HHR.MS.MRG.3IMERG.20230101-S000000-E002959.0000.V07B.HDF5
        name = url. \
            replace(
            'https://data.gesdisc.earthdata.nasa.gov/data/GPM_L3/GPM_3IMERGHH.07/2023/001/3B-HHR.MS.MRG.3IMERG.',
            '').replace('.0000.V07B.HDF5', '.HDF5')
        size = 0  # 初始化已下载大小
        chunk_size = 1024  # 每次下载的数据大小
        content_size = int(response.headers['content-length'])  # 下载文件总大小
        if response.status_code == 200:  # 判断是否响应成功
            print('Start download,[File size]:{size:.2f} MB'.format(
                size=content_size / chunk_size / 1024))  # 开始下载，显示下载文件大小
            filepath = rf'{path}\{name}'  # 设置图片name，注：必须加上扩展名
            with open(filepath, 'wb') as file:  # 显示进度条
                for data in response.iter_content(chunk_size=chunk_size):
                    file.write(data)
                    size += len(data)
                    print('\r' + '[下载进度]:%s%.2f%%' % (
                        '>' * int(size * 50 / content_size), float(size / content_size * 100)), end=' ')
        end = time.time()  # 下载结束时间
        print('Download completed!,times: %.2f秒' % (end - start))  # 输出下载用时时间

    def simulate(self, url, driver, path):
        names = path.split('/')[-1].split('.')
        date = names[4].split('-')[0]
        move_path = path.replace('.'.join(names[:4]), f'{date}{os.path.sep}{".".join(names[:4])}')
        if os.path.exists(move_path):
            print(f'{move_path} already exists,skip download')
            return
        start = time.time()  # 下载开始时间
        if not os.path.exists(path):
            driver.get(url)
        while not os.path.exists(path):
            time.sleep(1)
        # 3B-HHR-E.MS.MRG.3IMERG.20230101-S033000-E035959.0210.V07B.HDF5
        if not os.path.exists(os.path.dirname(move_path)):
            os.makedirs(os.path.dirname(move_path))
        shutil.move(path, move_path)
        end = time.time()  # 下载结束时间
        print('Download completed!,times: %.2f秒' % (end - start))  # 输出下载用时时间

    def start_download(self):
        download_file_path = self.download_file_path
        oprint(f'开始下载:{download_file_path}')
        f = open(download_file_path, 'r')
        if not os.path.exists('downloaded.txt'):
            open('downloaded.txt', 'w').close()
        save_f = open('downloaded.txt', 'r+')

        downloaded_list = save_f.readlines()
        downloaded_list = [u.replace('\n', '') for u in downloaded_list if u.count('.pdf') == 0]
        save_f.close()
        with open('downloaded.txt', 'a') as save_f_w:
            driver = self.make_driver('https://urs.earthdata.nasa.gov', download_dir=self.save_path, is_headless=True)
            time.sleep(10)
            username = driver.find_element(By.ID, 'username')
            passwd = driver.find_element(By.ID, 'password')
            username.send_keys(self.username)
            passwd.send_keys(self.passwd)
            driver.find_element(By.NAME, 'commit').click()
            time.sleep(5)
            for line in f.readlines():
                url = line.strip()
                if line.replace('\n', '') not in downloaded_list:
                    oprint(url)
                    file_path = self.save_path + os.sep + url.split('/')[-1]
                    try:
                        self.simulate(url, driver, file_path)
                        save_f_w.write(url + '\n')
                    except Exception as e:
                        print(f'{url} download failed')
                else:
                    print(f'{url} already downloaded')
            driver.close()


def getNPfromHDF(hdf_path, lock=None, save_file=True):
    if lock:
        lock.acquire()
    ds = nc.Dataset(hdf_path)
    if lock:
        lock.release()
    np_data = np.array(ds.groups['Grid']['precipitation'][:]).astype(np.float32)
    np_data = np_data[np.newaxis, :]
    ds.close()
    np_data = rearrange(np_data[0][0], 'w h->h w')[::-1, :]
    np_data[np_data < 0] = 0
    if save_file:
        np2tif(np_data, save_path='np2tif_dir', left=-180, top=90, x_res=0.1, y_res=0.1, out_name='IMERG',
               dtype='float32')
    return np_data



import time
from jacksung.utils.log import oprint as print


class NoFileException(Exception):
    def __init__(self, file_name):
        self.file_name = file_name
        super().__init__(f'No such file: {file_name}')


class NanNPException(Exception):
    def __init__(self, file_name):
        self.file_name = file_name
        super().__init__(f'Nan value in np data: {file_name}')


def wait_fun(fun, args, catch_exception=Exception, sleep_time=0.5, wait_time=5, open_log=True):
    try:
        return fun(*args)
    except catch_exception as e:
        if open_log:
            print(f'Task {args} failed, retry in {sleep_time}s, remain waiting time: {wait_time}s')
        if wait_time <= 0:
            raise e
        else:
            time.sleep(sleep_time)
            return wait_fun(fun, args, catch_exception, sleep_time=sleep_time, wait_time=wait_time - sleep_time)



import numpy as np


def mean_std_part2all(num_list, mean_list, std_list):
    # [(n,mean,std)]
    group_num = len(num_list)
    dim = mean_list[0].shape
    data_num = sum(num_list)

    global_mean = np.zeros(dim, dtype=np.float64)
    global_var = np.zeros(dim, dtype=np.float64)
    for i in range(group_num):
        global_mean += (num_list[i] / data_num * mean_list[i])
    for i in range(group_num):
        global_var += (num_list[i] / data_num * (std_list[i] ** 2 + (global_mean - mean_list[i]) ** 2))
    return global_mean, np.sqrt(global_var)


def cal_mean_std_one_loop(s, ss, count):
    mean_pixel = s / count
    mean_level = mean_pixel.mean((-2, -1))

    var_pixel = ss / count - np.square(mean_pixel)
    ss_level = ss.mean((-2, -1))
    var_level = ss_level / count - np.square(mean_level)

    std_level = np.sqrt(var_level)
    std_pixel = np.sqrt(var_pixel)

    return mean_pixel, std_pixel, mean_level, std_level


if __name__ == '__main__':
    d1 = np.random.rand(2, 3, 2, 2)
    d2 = np.random.rand(3, 3, 2, 2)
    d = np.concatenate([d1, d2])
    s1 = d1.sum((0,))
    ss1 = np.array([i ** 2 for i in d1]).sum((0,))

    s2 = d2.sum((0,))
    ss2 = np.array([i ** 2 for i in d2]).sum((0,))

    mean_pixel1, std_pixel1, mean_level1, std_level1 = cal_mean_std_one_loop(s1, ss1, 2)
    mean_pixel2, std_pixel2, mean_level2, std_level2 = cal_mean_std_one_loop(s2, ss2, 3)
    # print('*' * 40)
    # print(mean_pixel1, end='\n\n')
    # print(std_pixel1, end='\n\n')
    # print(mean_level1, end='\n\n')
    print(std_level1)
    # print('*' * 40)
    # print(mean_pixel2, end='\n\n')
    # print(std_pixel2, end='\n\n')
    # print(mean_level2, end='\n\n')
    print(std_level2)

    mean_pixel, std_pixel = mean_std_part2all([2, 3], [mean_pixel1, mean_pixel2], [std_pixel1, std_pixel2])
    mean_level, std_level = mean_std_part2all([2, 3], [mean_level1, mean_level2], [std_level1, std_level2])
    # print('*' * 40)
    # print(mean_pixel, end='\n\n')
    # print(std_pixel, end='\n\n')
    # print(mean_level, end='\n\n')
    print(std_level)
    print('*' * 40)
    print(d1.std((0, -2, -1)))
    print(d2.std((0, -2, -1)))
    print(d.std((0, -2, -1)))



import torch.nn as nn
from jacksung.ai.GeoNet.m_block import FEB, Tail, Head, DownBlock, UpBlock, CubeEmbeding, CubeUnEmbeding, Norm, ACT
import torch


class GeoNet(nn.Module):
    def __init__(self, window_sizes, n_lgab, c_in, c_lgan, r_expand=4, down_sample=4, num_heads=8, task='pred',
                 downstage=2):
        super(GeoNet, self).__init__()
        self.window_sizes = window_sizes
        self.n_lgab = n_lgab
        self.c_in = c_in
        self.c_lgan = c_lgan
        self.r_expand = r_expand
        self.task = task
        self.down_sample = down_sample
        # define head module
        if self.task == 'prec':
            self.head = Head(self.c_lgan, self.c_in, self.down_sample)
        else:
            self.head = CubeEmbeding(self.c_lgan, self.c_in, self.down_sample)
            self.head_res = Head(self.c_lgan, self.c_in, self.down_sample)
        # self.head = Head(self.c_lgan, self.c_in * 2, self.down_sample)
        # define body module
        self.body = nn.ModuleList()
        self.downstage = downstage
        for i in range(self.n_lgab):
            if i / self.downstage in [1]:
                self.body.append(DownBlock(self.c_lgan, 2))
                self.c_lgan = self.c_lgan * 2
            elif (self.n_lgab - i) / self.downstage in [1]:
                self.body.append(UpBlock(self.c_lgan // 2, 2))
                self.c_lgan = self.c_lgan // 2
            self.body.append(
                FEB(self.c_lgan, self.r_expand, self.window_sizes[i % len(self.window_sizes)], num_heads=num_heads))
        # self.conv1 = nn.Conv2d(self.c_lgan, self.c_lgan, 3, 1, 1)
        # self.norm = Norm(self.c_lgan)
        # self.act = ACT()
        # self.conv2 = nn.Conv2d(self.c_lgan, self.c_lgan, 3, 1, 1)
        self.tail = Tail(self.c_lgan, 1 if self.task == 'prec' else self.c_in, down_sample)
        # self.tail = nn.ConvTranspose3d(self.c_lgan, 5 if self.task == 'prec' else self.c_in, kernel_size=(2, 2, 2),
        #                                stride=(2, 2, 2))
        # self.tail = CubeUnEmbeding(self.c_lgan, 5 if self.task == 'prec' else self.c_in, self.down_sample)

    def forward(self, f, x, roll=0):
        # head
        head_res = None
        if self.task == 'pred':
            head_res = self.head_res(x)
            x = torch.stack([x, f], dim=2)
        if roll > 0:
            x = torch.roll(x, shifts=roll, dims=-1)
        # f, x = rearrange(x, 'b c z h w->b (c z) h w'), rearrange(f, 'b c z h w->b (c z) h w')
        # x = nn.functional.interpolate(x, size=[1600, 2000], mode='bilinear')
        # f = nn.functional.interpolate(f, size=[1600, 2000], mode='bilinear')
        x = self.head(x)
        # shortcut = x
        # body
        x_res = None
        for idx, stage in enumerate(self.body):
            if idx / self.downstage in [1]:
                x_res = x
            x = stage(x)
            if (self.n_lgab + 1 - idx) / self.downstage in [1]:
                x += x_res
        if self.task == 'pred':
            x = head_res + x
        # tail
        x = self.tail(x)
        # x = x[:, :, 0, :, :]
        if roll > 0:
            x = torch.roll(x, shifts=-roll, dims=-1)
        # x = nn.functional.interpolate(x, size=[1607, 2008], mode='bilinear')
        return x

    def init_model(self):
        print('Initializing the model!')
        for m in self.children():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.xavier_normal_(m.weight)

    def load(self, state_dict, strict=True):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            name = name[name.index('.') + 1:]
            if name in own_state.keys():
                if isinstance(param, nn.Parameter):
                    param = param.data
                try:
                    own_state[name].copy_(param)
                    # own_state[name].requires_grad = False
                except Exception as e:
                    err_log = f'While copying the parameter named {name}, ' \
                              f'whose dimensions in the model are {own_state[name].size()} and ' \
                              f'whose dimensions in the checkpoint are {param.size()}.'
                    if not strict:
                        print(err_log)
                    else:
                        raise Exception(err_log)
            elif strict:
                raise KeyError(f'unexpected key {name} in {own_state.keys()}')
            else:
                print(f'{name} not loaded by model')


if __name__ == '__main__':
    pass



import concurrent.futures
import threading
import multiprocessing
import time
from tqdm import tqdm
import traceback

type_thread = 'type_thread'
type_process = 'type_process'


class ThreadingLock:
    def __init__(self):
        self.lock = threading.Lock()
        self.owner = None

    def acquire(self, blocking=True):
        acquired = self.lock.acquire(blocking)
        if acquired:
            self.owner = threading.current_thread().name
        return acquired

    def release(self):
        self.lock.release()
        self.owner = None

    def get_owner(self):
        return self.owner


def init_process(pl):
    """进程池初始化函数，只设置进程锁"""
    global p_lock
    p_lock = pl


def init_thread(tl):
    """线程池初始化函数，只设置线程锁"""
    global t_lock
    t_lock = tl


class MultiTasks:
    def __init__(self, threads=10, pool=type_thread, desc="Mult. Pro."):
        self.threads = threads
        self.task_list = {}
        self.features = {}
        self.results = {}
        self.submitted = []
        self.pool_type = pool

        if pool == type_thread:
            # 线程池不需要在初始化时传递锁
            self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=self.threads)
        elif pool == type_process:
            # 进程池需要特殊处理锁
            self.process_mutex = multiprocessing.Manager().Lock()
            self.executor = concurrent.futures.ProcessPoolExecutor(
                max_workers=self.threads,
                initializer=init_process,
                initargs=(self.process_mutex,)
            )
        self.desc = desc
        self.progress_bar = None

    def add_task(self, k, function, args):
        self.task_list[k] = (function, args)

    def execute_task_nowait(self, save=False, print_log=False):
        for k, f_and_a in self.task_list.items():
            count = 0
            if k not in self.submitted:
                r = self.executor.submit(f_and_a[0], *f_and_a[1])
                if save:
                    self.features[k] = r
                self.submitted.append(k)
                count += 1
            if print_log:
                print(f'Submitted tasks: {count}, Total submitted: {len(self.submitted)}/{len(self.task_list)}')

    def wrap_fun(self, fun, args):
        """包装函数，处理进度条更新"""
        try:
            result = fun(*args)
        except Exception as e:
            traceback.print_exc()
            result = e
        if self.progress_bar:
            # 使用适当的方式更新进度条
            if self.pool_type == type_thread:
                # 线程环境下直接更新
                self.progress_bar.update(1)
            else:
                # 进程环境下需要特殊处理，这里简化处理
                raise Exception('进程环境下需要特殊处理，这里简化处理.')
        return result

    def execute_task(self, print_percent=True, desc=None):
        if print_percent:
            self.progress_bar = tqdm(total=len(self.task_list), desc=desc if desc else self.desc)

        # 根据池类型选择不同的执行策略
        if self.pool_type == type_thread:
            # 线程池直接执行
            for k, f_and_a in self.task_list.items():
                if k not in self.submitted:
                    self.features[k] = self.executor.submit(self.wrap_fun, f_and_a[0], f_and_a[1])
                    self.submitted.append(k)
        else:
            # 进程池需要避免传递不可序列化对象
            for k, f_and_a in self.task_list.items():
                if k not in self.submitted:
                    # 直接提交函数，不包装进度条更新
                    self.features[k] = self.executor.submit(f_and_a[0], *f_and_a[1])
                    self.submitted.append(k)

        # 收集结果并更新进度条
        for i, (k, feature) in enumerate(self.features.items()):
            self.results[k] = feature.result()
            if self.progress_bar and self.pool_type == type_process:
                # 进程环境下在主线程更新进度条
                self.progress_bar.update(1)

        if self.progress_bar:
            self.progress_bar.close()

        return self.results


def worker(i):
    # print(i)
    time.sleep(0.1)
    return i * 2


if __name__ == '__main__':
    mt = MultiTasks(10, pool=type_thread)
    for i in range(100):
        mt.add_task(i, worker, [i])
    mt.execute_task()



import shutil

from setuptools import setup, find_packages
from tinydb import TinyDB, Query
import subprocess

shutil.rmtree('build', ignore_errors=True)
shutil.rmtree('dist', ignore_errors=True)
shutil.rmtree('jacksung.egg-info', ignore_errors=True)
db = TinyDB('loacaldb.json')
infos = db.all()
if len(infos) == 0:
    version = '0.0.0.0'
    db.insert({'version': version})
else:
    version = db.all()[0]['version']
version = '.'.join(version.split('.')[:-1]) + '.' + str(int(version.split('.')[-1]) + 1)
db.update({'version': version})
db.close()
setup(
    name='jacksung',
    version=version,
    author='Zijiang Song',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'tqdm',
        'requests',
        'pymysql',
        'pytz',
        'selenium',
        'termcolor',
        'einops',
        'rasterio',
        'netCDF4',
        'pyyaml',
        'opencv-python',
        'Pillow',
        'openai',
        'satpy',
        'pyresample'
    ],
    entry_points={
        'console_scripts': [
            'ecnu_login = jacksung.utils.login:main',
            'watch_gpu = jacksung.utils.nvidia:main'
        ]
    }
)
try:
    subprocess.run(["git", "commit", "-am", rf"Update package {version}"], check=True)
except subprocess.CalledProcessError as e:
    print("Git 命令执行失败:", e)



import threading
import time


class Cache:
    def __init__(self, cache_len=120):
        self.cache_L = threading.Lock()
        self.cache = {}
        self.cache_list = []
        self.cache_len = cache_len

    # 判断key是否在缓存中，在则返回key对应的值，否则返回False, 同时锁住该key的所有查询，直到被放入数据
    def get_key_in_cache(self, key):
        try:
            while True:
                self.cache_L.acquire()
                if key in self.cache.keys():
                    if self.cache[key].is_ok:
                        result = self.cache[key].value
                        self.cache_list.remove(key)
                        self.cache_list.append(key)
                        self.cache_L.release()
                        return result
                else:
                    break
                self.cache_L.release()
                time.sleep(0.5)
            self.__add_key(key)
            self.cache_L.release()
            return None
        except Exception as e:
            self.cache_L.release()
            raise e

    def add_key(self, key, value):
        try:
            self.cache_L.acquire()
            if key not in self.cache.keys():
                self.__add_key(key)
            self.cache[key].set_value(value)
            self.cache_L.release()
            return value
        except Exception as e:
            self.cache_L.release()
            raise e

    def __add_key(self, key):
        self.cache_list.append(key)
        self.cache[key] = self.CacheClass(key)
        if len(self.cache_list) > self.cache_len:
            del_key = self.cache_list.pop(0)
            del self.cache[del_key]

    class CacheClass:
        def __init__(self, key):
            self.is_ok = False
            self.key = key
            self.value = None

        def set_value(self, value):
            self.value = value
            self.is_ok = True


if __name__ == '__main__':
    cache = Cache(2)
    cache.add_key('a', 1)
    cache.add_key('b', 2)
    print(cache.get_key_in_cache('a'))
    cache.add_key('c', 3)
    print(cache)



import os.path
import random
from datetime import timedelta, datetime

import torch
from jacksung.utils.time import RemainTime, Stopwatch, cur_timestamp_str
from jacksung.ai.utils.norm_util import PredNormalization, PrecNormalization, Normalization
import numpy as np
from jacksung.utils.data_convert import np2tif, fill_nan_with_window_mean
from jacksung.utils.cache import Cache
from jacksung.ai.utils import fy, goes, metsat
from einops import rearrange
from jacksung.ai.utils.util import parse_config, data_to_device, clipSatelliteNP
from jacksung.ai.GeoNet.m_network import GeoNet
from jacksung.ai.GeoNet.m_networkV2 import GeoNet as GeoNetV2
from jacksung.utils.exception import NoFileException, NanNPException
import torch.nn as nn

AGRI = 'agri'
ABI = 'abi'
SEVIRI = 'seviri'
AHI = 'ahi'
FCI = 'fci'


class GeoAttX:
    def __init__(self, norm_path=None, config=None, root_path=None, task_type=None, area=((100, 140, 10), (20, 60, 10)),
                 lock=None, print_timelog=False):
        self.root_path = None
        self.timestamp = None
        self.dir_name = None
        self.print_timelog = print_timelog
        self.norm_path = norm_path
        self.device, self.args = parse_config(config)
        self.task_type = task_type
        self.set_root_path(root_path)
        self.area = area
        self.lock = lock

    def get_root_path(self):
        return self.root_path

    def get_dir_name(self):
        return self.dir_name

    def set_root_path(self, root_path=None, dir_name=None):
        self.timestamp = cur_timestamp_str()
        root_path = root_path if root_path else self.args.save_path
        dir_name = dir_name if dir_name else self.task_type + '-' + self.args.model + '-' + self.timestamp + '_' + str(
            random.randint(1000, 9999))
        self.dir_name = dir_name
        self.root_path = os.path.join(root_path, dir_name)

    def load_model(self, path, version=1, c_in=None):
        if version == 1:
            model = GeoNet(window_sizes=self.args.window_sizes, n_lgab=self.args.n_lgab, c_in=self.args.c_in,
                           c_lgan=self.args.c_lgan, r_expand=self.args.r_expand, down_sample=self.args.down_sample,
                           num_heads=self.args.num_heads, task=self.task_type if self.task_type else self.args.task,
                           downstage=self.args.downstage)
        else:
            model = GeoNetV2(window_sizes=self.args.window_sizes, n_lgab=self.args.n_lgab, c_in=c_in,
                             c_lgan=self.args.c_lgan, r_expand=self.args.r_expand, down_sample=self.args.down_sample,
                             num_heads=self.args.num_heads, downstage=self.args.downstage)

        ckpt = torch.load(path, map_location=torch.device(self.device))
        model.load(ckpt['model_state_dict'])
        model = model.to(self.device)
        model = model.eval()
        model.requires_grad_(False)
        return model


class GeoAttX_I(GeoAttX):
    def __init__(self, norm_path, x1_path, x4_path, x12_path, root_path=None, config='config_predict.yml',
                 area=((100, 140, 10), (20, 60, 10)), cache_size=1, lock=None, device=None):
        super().__init__(norm_path=norm_path, config=config, root_path=root_path, task_type='pred', area=area,
                         lock=lock)
        if device is not None:
            self.device = device
        self.f, self.n, self.ys = None, None, None
        self.x1 = self.load_model(x1_path)
        self.x4 = self.load_model(x4_path)
        self.x12 = self.load_model(x12_path)
        # self.x48 = self.load_model(x48_path)
        self.norm = PredNormalization(self.norm_path)
        self.norm.mean, self.norm.std = data_to_device([self.norm.mean, self.norm.std], self.device, self.args.fp)
        self.ld = None
        self.cache = Cache(cache_size)

    def save(self, current_date, ys, save_target=True):
        start_str = current_date.strftime("%Y%m%d%H%M%S")
        end_str = (current_date + timedelta(minutes=14, seconds=59)).strftime("%Y%m%d%H%M%S")
        file_name = rf'FY4B-_AGRI--_N_DISK_{105 if current_date > datetime(year=2023, month=3, day=1) else 133}' + \
                    rf'0E_L1-_FDI-_MULT_NOM_{start_str}_{end_str}_4000M_V0001.HDF'
        file_info = fy.prase_filename(file_name)

        for idx, (k, y) in enumerate(ys.items()):
            # coord = getFY_coord_min(ld)
            coord = fy.getFY_coord_clip(self.area)
            np2tif(y, save_path=self.root_path, out_name=f'{k.strftime("%Y%m%d_%H%M%S")}', coord=coord,
                   dtype=np.float32, print_log=False, dim_value=[{'value': [str(x) for x in list(range(9, 16))]}])
            td = k - file_info['start']
            mins = td.days * 24 * 60 + td.seconds // 60
            target_filename = self.get_filename(file_name, mins)
            p_path = self.get_path_by_filename(target_filename)
            if idx >= 1 and os.path.exists(p_path) and save_target:
                coord = fy.getFY_coord_clip(self.area)
                p_data = fy.getNPfromHDF(p_path)
                p_data = clipSatelliteNP(p_data, ld=self.ld, area=self.area)
                if p_data is not None:
                    p_data = p_data[2:, :, :]
                    np2tif(p_data, save_path=self.root_path, out_name=f'target_{k.strftime("%Y%m%d_%H%M%S")}',
                           print_log=False, coord=coord, dtype=np.float32,
                           dim_value=[{'value': [str(x) for x in list(range(9, 16))]}])
        print(f'data saved in {self.root_path}')
        with open(os.path.join(self.root_path, 'info.log'), 'a') as f:
            f.write(f'输入数据：{file_info["start"]} {file_info["position"]} {file_info["end"]}\n')
            for k, y in ys.items():
                f.write(f'预测：{k}\n')
        return self.root_path

    def get_filename(self, file_name, mins):
        file_info = fy.prase_filename(file_name)
        new_file_name = file_name.replace(
            f'{file_info["start"].strftime("%Y%m%d%H%M%S")}_{file_info["end"].strftime("%Y%m%d%H%M%S")}',
            f'{(file_info["start"] + timedelta(minutes=mins)).strftime("%Y%m%d%H%M%S")}_'
            f'{(file_info["end"] + timedelta(minutes=mins)).strftime("%Y%m%d%H%M%S")}')
        return new_file_name

    def get_path_by_filename(self, file_name):
        file_info = fy.prase_filename(file_name)
        return f'{self.args.pred_data_path}{os.sep}{file_info["start"].year}{os.sep}{file_info["start"].month}{os.sep}{file_info["start"].day}{os.sep}{file_name}'
        # return f'{self.data_path}/{file_name}'

    def numpy2tensor(self, f_data):
        f_data = torch.from_numpy(f_data)
        f = data_to_device([f_data], self.device, self.args.fp)[0]
        f = rearrange(f, '(b c) h w -> b c h w', b=1)
        return f

    def get_exist_by_filename_and_mins(self, file_name, mins):
        f_path = self.get_filename(file_name, mins)
        f_path = self.get_path_by_filename(f_path)
        if not os.path.exists(f_path):
            raise NoFileException(f_path)
        f_data = fy.getNPfromHDF(f_path)
        if f_data is None:
            raise NanNPException(f_path)
        f_data = clipSatelliteNP(f_data, ld=self.ld, area=self.area)
        if type(f_data) is not str:
            f_data = f_data[2:, :, :]
        else:
            raise NanNPException(f_path)
        # f_data = zoom(f_data.astype(np.float32), (1, 1 / 5, 1 / 5))
        return self.numpy2tensor(f_data)

    def mean_std2Tensor(self, in_data, h, w):
        in_data = rearrange(in_data, '(b h w) c->b c h w', h=1, w=1)
        in_data = in_data.expand(1, 7, h, w)
        return in_data

    def secdOrderStd(self, t_data, mean0, std0, mean, std):
        h, w = t_data.shape[2], t_data.shape[3]
        t_data = (t_data - self.mean_std2Tensor(mean0, h, w)) / self.mean_std2Tensor(std0, h, w)
        t_data = t_data * self.mean_std2Tensor(std, h, w) + self.mean_std2Tensor(mean, h, w)
        return t_data

    def predict(self, current_date, step=360, p_steps=(48, 12, 4, 1), print_log=True):
        try:
            start_str = current_date.strftime("%Y%m%d%H%M%S")
            end_str = (current_date + timedelta(minutes=14, seconds=59)).strftime("%Y%m%d%H%M%S")
            file_name = rf'FY4B-_AGRI--_N_DISK_{105 if current_date > datetime(year=2023, month=3, day=1) else 133}' + \
                        rf'0E_L1-_FDI-_MULT_NOM_{start_str}_{end_str}_4000M_V0001.HDF'
            file_info = fy.prase_filename(file_name)
            self.ld = int(file_info["position"])
            step = step // 15
            # if step > 24:
            #     step = 24
            if print_log:
                print(f'当前时刻:{file_info["start"]}\n预测长度:{step * 15}分钟')
            task_progress = []
            p_steps = sorted(p_steps, reverse=True)
            while step > 0:
                for p_step in p_steps:
                    if step >= p_step:
                        task_progress.append(p_step)
                        step -= p_step
                        break
            task_progress.reverse()
            if print_log:
                print(f'正在预测:{file_info["start"] + timedelta(minutes=sum(task_progress) * 15)}...')
            n = self.get_exist_by_filename_and_mins(file_name, 0)
            now_date = file_info["start"]
            porcess_list = {now_date: n.detach().cpu().numpy()[0]}
            n = self.norm.norm(n)
            o_mean = torch.mean(n, dim=(2, 3))
            o_std = torch.std(n, dim=(2, 3))
            for step in task_progress:
                pre_date = now_date - timedelta(minutes=15 * step)
                now_date += timedelta(minutes=15 * step)
                if print_log:
                    print(f'正在预测 {now_date}...')
                if pre_date in porcess_list:
                    f = porcess_list[pre_date]
                    f = self.numpy2tensor(f)
                else:
                    f = self.get_exist_by_filename_and_mins(file_name,
                                                            -int((file_info["start"] - pre_date).seconds / 60))
                f = self.norm.norm(f)
                st = Stopwatch()
                y_ = eval(f'self.x{step}(f, n)')
                if print_log:
                    print(f'预测耗时: {st.reset()} 秒')
                # mean and std再标准化
                y_mean = torch.mean(y_, dim=(2, 3))
                y_std = torch.std(y_, dim=(2, 3))
                # 二次标准化
                y_ = self.secdOrderStd(y_, y_mean, y_std, o_mean, o_std)
                # y_ = torch.Tensor(np.zeros((1, 7, 2400, 2400), dtype=np.float32)).to(self.device)
                # for i in range(5):
                #     for j in range(5):
                #         n_ = n[:, :, 480 * i:480 * (i + 1), 480 * j:480 * (j + 1)]
                #         f_ = f[:, :, 480 * i:480 * (i + 1), 480 * j:480 * (j + 1)]
                #         y_[:, :, 480 * i:480 * (i + 1), 480 * j:480 * (j + 1)] = eval(f'self.x{step}(f_, n_)')
                del f
                n = y_
                porcess_list[now_date] = self.norm.denorm(y_).detach().cpu().numpy()[0]
        except (NoFileException, NanNPException) as e:
            os.makedirs(self.root_path, exist_ok=True)
            with open(os.path.join(self.root_path, 'err.log'), 'a') as f:
                filename = e.file_name.split(os.sep)[-1]
                file_info = fy.prase_filename(filename)
                f.write(f'{e.__class__}: {file_info["start"]} {e.file_name}\n')
            return {}
        return porcess_list


class GeoAttX_P(GeoAttX):
    def __init__(self, norm_path=None, model_path=None, root_path=None, config='predict_qpe.yml',
                 area=((100, 140, 10), (20, 60, 10)),
                 device=None):
        super().__init__(norm_path=norm_path, config=config, root_path=root_path, task_type='prec', area=area)
        if device is not None:
            self.device = device
        self.model = self.load_model(model_path)
        self.norm = PrecNormalization(self.norm_path)
        self.norm.mean_fy, self.norm.mean_qpe, self.norm.std_fy, self.norm.std_qpe = \
            data_to_device([self.norm.mean_fy, self.norm.mean_qpe, self.norm.std_fy, self.norm.std_qpe],
                           self.device, self.args.fp)

    def save(self, y, save_name, info_log=True, print_log=True):
        np2tif(y, save_path=self.root_path, out_name=save_name, coord=fy.getFY_coord_clip(self.area), dtype=np.float32,
               print_log=False, dim_value=[{'value': ['qpe']}])
        if print_log:
            print(f'data saved in {self.root_path}')
        if info_log:
            with open(os.path.join(self.root_path, 'info.log'), 'a') as f:
                f.write(f'QPE 反演：{save_name}\n')
        return self.root_path

    def predict(self, fy_npy):
        try:
            if type(fy_npy) is str:
                print(f'正在反演:{fy_npy}...')
                if not os.path.exists(fy_npy):
                    raise NoFileException(fy_npy)
                n_data = np.load(fy_npy)
            elif type(fy_npy) is np.ndarray:
                n_data = fy_npy
            else:
                raise Exception('输入数据类型错误，仅支持文件路径或numpy数组')
            n_data = torch.from_numpy(n_data)
            n_data = data_to_device([n_data], self.device, self.args.fp)[0]
            n_data = rearrange(n_data, '(b t c) h w -> b t c h w', b=1, t=1)
            n = self.norm.norm(n_data, norm_type='fy')[:, 0, :, :, :]
            y_ = self.model(n, n)
            y = self.norm.denorm(y_, norm_type='qpe').detach().cpu().numpy()[0]
            return y
        except NoFileException as e:
            os.makedirs(self.root_path, exist_ok=True)
            with open(os.path.join(self.root_path, 'err.log'), 'a') as f:
                filename = e.file_name.split(os.sep)[-1]
                file_info = fy.prase_filename(filename)
                f.write(f'Not exist {file_info["start"]} {e.file_name}\n')
            return None


class Huayu(GeoAttX):
    def __init__(self, norm_path=None, model_path=None, root_path=None, config='config.yml',
                 area=((100, 140, 10), (20, 60, 10)), device=None, print_timelog=False):
        super().__init__(norm_path=norm_path, config=config, root_path=root_path, task_type='prem', area=area,
                         print_timelog=print_timelog)
        if device is not None:
            self.device = device
        if self.args.sensor_type == AGRI:
            self.sensor = AGRI
            self.satellite_num = 2
            self.satellite_channel = 7
        elif self.args.sensor_type == ABI:
            self.sensor = ABI
            self.satellite_num = 3
            self.satellite_channel = 9
        elif self.args.sensor_type == SEVIRI:
            self.sensor = SEVIRI
            self.satellite_num = 2
            self.satellite_channel = 7
        else:
            raise ValueError(f'Unknown sensor type{self.args.sensor_type}')
        st = Stopwatch()
        mean_std_npy = np.load(os.path.join(rf'{self.norm_path}', 'mean_std.npy'))
        self.model = self.load_model(model_path, version=2, c_in=self.satellite_channel)
        self.satellite_norm = Normalization(mean_std_npy, (0, self.satellite_channel))
        self.imerg_norm = Normalization(mean_std_npy, (self.satellite_channel * self.satellite_num,
                                                       self.satellite_channel * self.satellite_num + 3))
        self.satellite_norm.mean, self.satellite_norm.std, self.imerg_norm.mean, self.imerg_norm.std = \
            data_to_device(
                [self.satellite_norm.mean, self.satellite_norm.std, self.imerg_norm.mean, self.imerg_norm.std],
                self.device, self.args.fp)
        if self.print_timelog:
            print(st.reset())

    def save(self, y, save_name, info_log=True, print_log=True):
        np2tif(y, save_path=self.root_path, out_name=save_name, coord=fy.getFY_coord_clip(self.area), dtype=np.float32,
               print_log=False, dim_value=[{'value': [self.sensor]}])
        if print_log:
            print(f'data saved in {self.root_path}')
        if info_log:
            with open(os.path.join(self.root_path, 'info.log'), 'a') as f:
                f.write(f'Huayu produced:{save_name}\n')
        return self.root_path

    def predict(self, npy_path=None, np_data=None, satellite_file=None, satellite=None, smooth=True, up=True, area=None,
                area_ij=None,
                satellite_date=None, fill_nan=False):
        try:
            st = Stopwatch()
            if npy_path is None and np_data is None:
                if self.sensor == AGRI:
                    n_data, coord = fy.getNPfromHDF(satellite_file, return_coord=True)
                    n_data = n_data[2:]
                elif self.sensor == ABI:
                    n_data, coord = goes.getNPfromDir(satellite_file, satellite_date, satellite=satellite,
                                                      return_coord=True)
                elif self.sensor == SEVIRI:
                    n_data, coord = metsat.getNPfromNAT(satellite_file, return_coord=True)
                else:
                    raise ValueError(f'Unknown sensor type{self.args.sensor_type}')
                if fill_nan:
                    n_data = fill_nan_with_window_mean(n_data, window_size=(9, 9))
                if area is not None:
                    self.area = area
                elif area_ij is not None:
                    self.area = \
                        ((coord.ld + (area_ij[1][0] - 1200) * 0.05, coord.ld + (area_ij[1][1] - 1200) * 0.05, 10),
                         (60 - area_ij[0][1] * 0.05, 60 - area_ij[0][0] * 0.05, 10))
                n_data = clipSatelliteNP(n_data, coord.ld, self.area)
            elif np_data is None:
                n_data = np.load(npy_path)
            else:
                n_data = np_data
            n_data = torch.from_numpy(n_data)
            n_data = data_to_device([n_data], self.device, self.args.fp)[0]
            n_data = rearrange(n_data, '(b c) h w -> b c h w', b=1)
            n = self.satellite_norm.norm(n_data)[:, :, :, :]
            ps = nn.PixelShuffle(2)
            ups = nn.PixelUnshuffle(2)
            if smooth:
                smooth = nn.AvgPool2d(kernel_size=3, stride=1, padding=1)
            n = ups(n)
            n = rearrange(n, 'b (c dsize) h w -> (b dsize) c h w', dsize=4)
            if not up:
                n = n.mean(dim=0, keepdim=True)
            if self.print_timelog:
                print(st.reset())
            y_ = self.model(n)
            if self.print_timelog:
                print(st.reset())
            y_ = rearrange(y_, '(b dsize) c h w -> b (c dsize) h w', b=1)
            if up:
                y_ = ps(y_)
            y = self.imerg_norm.denorm(y_)[0]
            y[0][y[1] > y[2]] = 0
            y[0][y[0] < 0] = 0
            y = rearrange(y[0], '(b h) w -> b h w', b=1)
            _, H, W = y.shape
            if smooth:
                y[0, 1:H - 1, 1:W - 1] = smooth(y)[0, 1:H - 1, 1:W - 1]
            if self.print_timelog:
                print(st.reset())
            return y.detach().cpu().numpy()
        except NoFileException as e:
            os.makedirs(self.root_path, exist_ok=True)
            with open(os.path.join(self.root_path, 'err.log'), 'a') as f:
                filename = e.file_name.split(os.sep)[-1]
                file_info = fy.prase_filename(filename)
                f.write(f'Not exist {file_info["start"]} {e.file_name}\n')
            return None



import time
from datetime import datetime
import pytz
from jacksung.utils.log import oprint
import calendar


def get_days_in_month(year, month):
    return calendar.monthrange(year, month)[1]


def cal_time(fun_str):
    # 记录第一个时间戳
    start_time = time.time()
    eval(fun_str)
    # 记录第二个时间戳
    end_time = time.time()
    # 计算耗时
    elapsed_time = end_time - start_time
    print("Elapsed Time:", elapsed_time, "seconds")

def get_time_str():
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
def cur_timestamp_str():
    now = datetime.now()
    year = str(now.year)
    month = str(now.month).zfill(2)
    day = str(now.day).zfill(2)
    hour = str(now.hour).zfill(2)
    minute = str(now.minute).zfill(2)
    second = str(now.second).zfill(2)
    microsecond = str(now.microsecond // 10000).zfill(2)
    content = "{}-{}{}-{}{}-{}{}".format(year, month, day, hour, minute, second, microsecond)
    return content


class RemainTime:
    def __init__(self, epoch):
        self.start_time = time.time()
        self.epoch = epoch
        self.now_epoch = 0

    def update(self, log_temp='Spe {:.0f}s, Rem Epoch:{}, Fin in {}', print_log=True, update_step=1):
        self.now_epoch += update_step
        epoch_time = time.time() - self.start_time
        epoch_remaining = self.epoch - self.now_epoch
        time_remaining = epoch_time * epoch_remaining
        pytz.timezone('Asia/Shanghai')  # 东八区
        t = datetime.fromtimestamp(int(time.time()) + time_remaining,
                                   pytz.timezone('Asia/Shanghai')).strftime('%Y-%m-%d %H:%M:%S')
        log = log_temp.format(epoch_time, epoch_remaining, t)
        if print_log:
            oprint(log)
        self.start_time = time.time()
        return epoch_remaining, t


class Stopwatch:
    def __init__(self):
        self.start_time = time.time()

    def reset(self, format_type='{:>5.1f}s'):
        timer_end = time.time()
        duration = timer_end - self.start_time
        self.start_time = timer_end
        return format_type.format(duration)

    def pinch(self, format_type='{:>5.1f}s'):
        return format_type.format(time.time() - self.start_time)


def getHumanSize(in_size):
    unit = 'B'
    if in_size >= 1024:
        in_size /= 1024
        unit = 'K'
    if in_size >= 1024:
        in_size /= 1024
        unit = 'M'
    if in_size >= 1024:
        in_size /= 1024
        unit = 'G'
    return f'{round(in_size, 2)} {unit}'


if __name__ == '__main__':
    print(getHumanSize(1023))
